from typing import Sequence, List, Optional

import numpy as np
import pandas as pd
import tensorflow as tf

from sknlp.vocab import Vocab
from .nlp_dataset import NLPDataset
from .utils import serialize_example


class ClassificationDataset(NLPDataset):

    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        is_multilabel: bool = True,
        max_length: Optional[int] = 80,
        text_segmenter: str = 'char'
    ):
        self.vocab = vocab
        self.labels = labels
        self.is_multilabel = is_multilabel
        self.max_length = max_length or 99999
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(df=df,
                         csv_file=csv_file,
                         in_memory=in_memory,
                         text_segmenter=text_segmenter,
                         text_dtype=tf.int32,
                         label_dtype=tf.float32,
                         text_padding_shape=(None,),
                         label_padding_shape=(None,),
                         text_padding_value=vocab[vocab.pad],
                         label_padding_value=0.0)

    def _text_transform(self, text: tf.Tensor) -> np.ndarray:
        tokens = super()._text_transform(text)
        return np.array(
            [self.vocab[t] for t in tokens[:self.max_length]], dtype=np.int32
        )

    def _label_binarizer(self, labels: List[str]) -> np.ndarray:
        label2idx = self.label2idx
        res = np.zeros(len(label2idx), dtype=np.float32)
        res[[label2idx[label] for label in labels if label in label2idx]] = 1
        return res

    def _label_transform(self, label: tf.Tensor) -> np.ndarray:
        label = super()._label_transform(label)
        if self.is_multilabel:
            labels = label.split('|')
        else:
            labels = [label]
        return self._label_binarizer(labels)

    def to_tfrecord(self, filename: str) -> None:

        def func(text: np.ndarray, label: np.ndarray):
            return tf.reshape(
                serialize_example(
                    (self.text_transform(text), self.label_transform(label)),
                    ('tensor', 'tensor')
                ), ()
            )

        tf_writer = tf.data.experimental.TFRecordWriter(filename)
        tf_writer.write(self._dataset.map(
            lambda t, l: tf.py_function(func, inp=[t, l], Tout=tf.string),
            num_parallel_calls=tf.data.experimental.AUTOTUNE
        ))

    @classmethod
    def from_tfrecord(cls, filename: str) -> tf.data.Dataset:

        def func(record: tf.Tensor):
            parsed_record = tf.io.parse_single_example(
                record,
                {
                    'feature0': tf.io.FixedLenFeature(
                        [], tf.string, default_value=''
                    ),
                    'feature1': tf.io.FixedLenFeature(
                        [], tf.string, default_value=''
                    )
                }
            )
            return (
                tf.io.parse_tensor(parsed_record['feature0'], tf.int32),
                tf.io.parse_tensor(parsed_record['feature1'], tf.float32)
            )

        return tf.data.TFRecordDataset(filename).map(func)



from typing import List

from tensorflow.keras.layers import Bidirectional, Dropout

from sknlp.layers import MLPLayer, LSTMP

from .supervised_model import SupervisedNLPModel


class TextRNN(SupervisedNLPModel):

    def __init__(self,
                 classes,
                 segmenter='jieba',
                 max_sequence_length=80,
                 embedding_size=100,
                 num_rnn_layers=1,
                 rnn_hidden_size=256,
                 rnn_projection_size=128,
                 rnn_recurrent_clip=3,
                 rnn_projection_clip=3,
                 rnn_input_dropout=0.5,
                 rnn_recurrent_dropout=0.5,
                 rnn_output_dropout=0.5,
                 rnn_last_connection=None,
                 num_fc_layers=2,
                 fc_hidden_size=256,
                 rnn_kernel_initializer='glorot_uniform',
                 rnn_recurrent_initializer='orthogonal',
                 rnn_projection_initializer='glorot_uniform',
                 rnn_bias_initializer='zeros',
                 rnn_kernel_regularizer=None,
                 rnn_recurrent_regularizer=None,
                 rnn_projection_regularizer=None,
                 rnn_bias_regularizer=None,
                 rnn_kernel_constraint=None,
                 rnn_recurrent_constraint=None,
                 rnn_projection_constraint=None,
                 rnn_bias_constraint=None,
                 text2vec=None,
                 **kwargs):
        super().__init__(classes,
                         segmenter=segmenter,
                         embedding_size=embedding_size,
                         max_sequence_length=max_sequence_length,
                         text2vec=text2vec,
                         **kwargs)
        self._num_rnn_layers = num_rnn_layers
        self._rnn_hidden_size = rnn_hidden_size
        self._rnn_last_connection = rnn_last_connection
        self._rnn_projection_size = rnn_projection_size
        self._rnn_recurrent_clip = rnn_recurrent_clip
        self._rnn_projection_clip = rnn_projection_clip
        self._rnn_input_dropout = rnn_input_dropout
        self._rnn_recurrent_dropout = rnn_recurrent_dropout
        self._rnn_output_dropout = rnn_output_dropout
        self._rnn_kernel_initializer = rnn_kernel_initializer
        self._rnn_recurrent_initializer = rnn_recurrent_initializer
        self._rnn_projection_initializer = rnn_projection_initializer
        self._rnn_bias_initializer = rnn_bias_initializer
        self._rnn_kernel_regularizer = rnn_kernel_regularizer
        self._rnn_recurrent_regularizer = rnn_recurrent_regularizer
        self._rnn_projection_regularizer = rnn_projection_regularizer
        self._rnn_bias_regularizer = rnn_bias_regularizer
        self._rnn_kernel_constraint = rnn_kernel_constraint
        self._rnn_recurrent_constraint = rnn_recurrent_constraint
        self._rnn_projection_constraint = rnn_projection_constraint
        self._rnn_bias_constraint = rnn_bias_constraint
        self._num_fc_layers = num_fc_layers
        self._fc_hidden_size = fc_hidden_size

    def build_encode_layer(self, inputs):
        for i in range(self._num_rnn_layers):
            return_sequences = (
                i != self._num_rnn_layers - 1
                or self._rnn_last_connection != 'last'
            )
            rnn_outputs = Bidirectional(LSTMP(
                self._rnn_hidden_size,
                self._rnn_projection_size,
                kernel_initializer=self._rnn_kernel_initializer,
                recurrent_initializer=self._rnn_recurrent_initializer,
                projection_initializer=self._rnn_projection_initializer,
                bias_initializer=self._rnn_bias_initializer,
                kernel_regularizer=self._rnn_kernel_regularizer,
                recurrent_regularizer=self._rnn_recurrent_regularizer,
                projection_regularizer=self._rnn_projection_regularizer,
                bias_regularizer=self._rnn_bias_regularizer,
                kernel_constraint=self._rnn_kernel_constraint,
                recurrent_constraint=self._rnn_recurrent_constraint,
                projection_constraint=self._rnn_projection_constraint,
                bias_constraint=self._rnn_bias_constraint,
                dropout=self._rnn_input_dropout,
                recurrent_dropout=self._rnn_recurrent_dropout,
                recurrent_clip=self._rnn_recurrent_clip,
                projection_clip=self._rnn_projection_clip,
                return_sequences=return_sequences
            ))(inputs)
            inputs = rnn_outputs
        if self._rnn_output_dropout:
            noise_shape = (None, 1, self._rnn_projection_size * 2)
            if self._rnn_last_connection == 'last':
                noise_shape = None
            return Dropout(self._rnn_output_dropout,
                           noise_shape=noise_shape,
                           name='rnn_output_dropout')(inputs)
        else:
            return inputs

    def build_output_layer(self, inputs):
        mlp = MLPLayer(self._num_fc_layers,
                       hidden_size=self._fc_hidden_size,
                       output_size=self._num_classes)
        return mlp(inputs)

    @property
    def output_names(self) -> List[str]:
        return ["mlp"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self._num_classes]]

    def get_config(self):
        return {
            **super().get_config(),
            'num_rnn_layers': self._num_rnn_layers,
            'rnn_hidden_size': self._rnn_hidden_size,
            'rnn_projection_size': self._rnn_projection_size,
            'rnn_recurrent_clip': self._rnn_recurrent_clip,
            'rnn_projection_clip': self._rnn_projection_clip,
            'rnn_input_dropout': self._rnn_input_dropout,
            'rnn_recurrent_dropout': self._rnn_recurrent_dropout,
            'rnn_output_dropout': self._rnn_output_dropout,
            'rnn_last_connection': self._rnn_last_connection,
            'num_fc_layers': self._num_fc_layers,
            'fc_hidden_size': self._fc_hidden_size,
            'rnn_kernel_initializer': self._rnn_kernel_initializer,
            'rnn_recurrent_initializer': self._rnn_recurrent_initializer,
            'rnn_projection_initializer': self._rnn_projection_initializer,
            'rnn_bias_initializer': self._rnn_bias_initializer,
            'rnn_kernel_regularizer': self._rnn_kernel_regularizer,
            'rnn_recurrent_regularizer': self._rnn_recurrent_regularizer,
            'rnn_projection_regularizer': self._rnn_projection_regularizer,
            'rnn_bias_regularizer': self._rnn_bias_regularizer,
            'rnn_kernel_constraint': self._rnn_kernel_constraint,
            'rnn_recurrent_constraint': self._rnn_recurrent_constraint,
            'rnn_projection_constraint': self._rnn_projection_constraint,
            'rnn_bias_constraint': self._rnn_bias_constraint
        }

    def get_custom_objects(self):
        return {**super().get_custom_objects(),
                'LSTMP': LSTMP,
                'MLPLayer': MLPLayer}



from typing import Sequence, List, Dict, Any, Optional, Union

from tensorflow.keras.layers import Dropout
import pandas as pd

from sknlp.layers import MLPLayer
from sknlp.vocab import Vocab
from sknlp.data import BertClassificationDataset
from ..text2vec import Bert2vec
from .deep_classifier import DeepClassifier


class BertClassifier(DeepClassifier):

    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        segmenter: str = None,
        embedding_size: int = 100,
        max_sequence_length: int = 100,
        num_fc_layers=2,
        fc_hidden_size=256,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ):
        super().__init__(classes,
                         is_multilabel=is_multilabel,
                         segmenter=segmenter,
                         embedding_size=embedding_size,
                         max_sequence_length=max_sequence_length,
                         text2vec=text2vec,
                         **kwargs)
        self._num_fc_layers = num_fc_layers
        self._fc_hidden_size = fc_hidden_size

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str]
    ) -> BertClassificationDataset:
        return BertClassificationDataset(
            vocab,
            list(labels),
            df=df,
            is_multilabel=self._is_multilabel,
            max_length=self._max_sequence_length,
            text_segmenter=segmenter
        )

    def get_config(self):
        return {
            **super().get_config(),
            'num_fc_layers': self._num_fc_layers,
            'fc_hidden_size': self._fc_hidden_size
        }

    def build_encode_layer(self, inputs):
        return Dropout(0.5)(inputs)

    def build_output_layer(self, inputs):
        mlp = MLPLayer(self._num_fc_layers,
                       hidden_size=self._fc_hidden_size,
                       output_size=self._num_classes)
        return mlp(inputs)

    @property
    def output_names(self) -> List[str]:
        return ["mlp"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self._num_classes]]

    def get_custom_objects(self):
        return {**super().get_custom_objects(), 'MLPLayer': MLPLayer}



from typing import Sequence, List, Dict, Any, Optional, Union

import numpy as np
import pandas as pd
from tabulate import tabulate
import tensorflow as tf

from sknlp.vocab import Vocab
from sknlp.data import ClassificationDataset
from sknlp.metrics import PrecisionWithLogits, RecallWithLogits
from sknlp.callbacks import FScore

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec

from .utils import (
    logits2probabilities,
    probabilities2classes,
    classification_fscore
)


class DeepClassifier(SupervisedNLPModel):

    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        segmenter: str = "jieba",
        embedding_size: int = 100,
        max_sequence_length: int = 100,
        text2vec: Optional[Text2vec] = None,
        **kwargs
    ):
        super().__init__(classes,
                         segmenter=segmenter,
                         embedding_size=embedding_size,
                         max_sequence_length=max_sequence_length,
                         text2vec=text2vec,
                         task="classification",
                         **kwargs)
        self._is_multilabel = is_multilabel

    def get_loss(self):
        if self._is_multilabel:
            return tf.keras.losses.BinaryCrossentropy(from_logits=True)
        else:
            return tf.keras.losses.CategoricalCrossentropy(from_logits=True)

    def get_callbacks(self) -> List[tf.keras.callbacks.Callback]:
        return [FScore()]

    def get_metrics(self) -> List[tf.keras.metrics.Metric]:
        if self._is_multilabel:
            return [PrecisionWithLogits(), RecallWithLogits()]
        else:
            return [PrecisionWithLogits(logits2scores="softmax"),
                    RecallWithLogits(logits2scores="softmax")]

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str]
    ) -> ClassificationDataset:
        return ClassificationDataset(
            vocab,
            list(labels),
            df=df,
            is_multilabel=self._is_multilabel,
            max_length=self._max_sequence_length,
            text_segmenter=segmenter
        )

    def predict_proba(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        batch_size: int = 128
    ) -> np.ndarray:
        logits = super().predict(X=X, dataset=dataset, batch_size=batch_size)
        return logits2probabilities(logits, self._is_multilabel)

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, List[float]] = 0.5,
        batch_size: int = 128
    ) -> Union[List[str], List[List[str]]]:
        probabilities = self.predict_proba(X=X, dataset=dataset, batch_size=batch_size)
        predictions = probabilities2classes(
            probabilities, self._is_multilabel, thresholds=thresholds
        )
        if self._is_multilabel:
            return [
                [self._idx2class[i] for i in prediction] for prediction in predictions
            ]
        else:
            return [self._idx2class[i] for i in predictions]

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, List[float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        predictions = self.predict(
            X=X, dataset=dataset, thresholds=thresholds, batch_size=batch_size
        )
        return classification_fscore(
            y, predictions, self._is_multilabel, classes=list(self._class2idx.keys())
        )

    def format_score(self, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "is_multilabel": self._is_multilabel}

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("algorithm", None)
        config.pop("task", None)
        return config

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "PrecisionWithLogits": PrecisionWithLogits,
            "RecallWithLogits": RecallWithLogits
        }



from typing import Optional

import tensorflow as tf
from tensorflow.keras.layers import Layer, Bidirectional, Dropout

from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint

from .lstmp import LSTMP


class MultiLSTMP(Layer):

    def __init__(
        self,
        num_layers: int,
        units: int,
        projection_size: int = 100,
        kernel_initializer: WeightInitializer = 'glorot_uniform',
        recurrent_initializer: WeightInitializer = 'orthogonal',
        projection_initializer: WeightInitializer = 'glorot_uniform',
        bias_initializer: WeightInitializer = 'zeros',
        kernel_regularizer: Optional[WeightRegularizer] = None,
        recurrent_regularizer: Optional[WeightRegularizer] = None,
        projection_regularizer: Optional[WeightRegularizer] = None,
        bias_regularizer: Optional[WeightRegularizer] = None,
        kernel_constraint: Optional[WeightConstraint] = None,
        recurrent_constraint: Optional[WeightConstraint] = None,
        projection_constraint: Optional[WeightConstraint] = None,
        bias_constraint: Optional[WeightConstraint] = None,
        input_dropout: float = 0.,
        recurrent_dropout: float = 0.,
        output_dropout: float = 0.,
        recurrent_clip: Optional[float] = None,
        projection_clip: Optional[float] = None,
        last_connection: str = 'last',
        **kwargs
    ):
        self.num_layers = num_layers
        self.units = units
        self.projection_size = projection_size
        self.kernel_initializer = kernel_initializer
        self.recurrent_initializer = recurrent_initializer
        self.projection_initializer = projection_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.recurrent_regularizer = recurrent_regularizer
        self.projection_regularizer = projection_regularizer
        self.bias_regularizer = bias_regularizer
        self.kernel_constraint = kernel_constraint
        self.recurrent_constraint = recurrent_constraint
        self.projection_constraint = projection_constraint
        self.bias_constraint = bias_constraint
        self.input_dropout = input_dropout
        self.recurrent_dropout = recurrent_dropout
        self.output_dropout = output_dropout
        self.recurrent_clip = recurrent_clip
        self.projection_clip = projection_clip
        self.last_connection = last_connection

        self.layers = []
        for i in range(self.num_layers):
            return_sequences = (
                i != self.num_layers - 1
                or self.last_connection != 'last'
            )
            self.layers.append(Bidirectional(LSTMP(
                self.units,
                self.projection_size,
                kernel_initializer=self.kernel_initializer,
                recurrent_initializer=self.recurrent_initializer,
                projection_initializer=self.projection_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                recurrent_regularizer=self.recurrent_regularizer,
                projection_regularizer=self.projection_regularizer,
                bias_regularizer=self.bias_regularizer,
                kernel_constraint=self.kernel_constraint,
                recurrent_constraint=self.recurrent_constraint,
                projection_constraint=self.projection_constraint,
                bias_constraint=self.bias_constraint,
                dropout=self.input_dropout,
                recurrent_dropout=self.recurrent_dropout,
                recurrent_clip=self.recurrent_clip,
                projection_clip=self.projection_clip,
                return_sequences=return_sequences
            )))
        noise_shape = (None, 1, self.projection_size * 2)
        if self.last_connection == 'last':
            noise_shape = None
        self.dropout_layer = Dropout(self.output_dropout, noise_shape=noise_shape)
        super().__init__(**kwargs)

    def call(self, inputs: int, initial_state: Optional[tf.Tensor] = None) -> tf.Tensor:
        for layer in self.layers:
            inputs = layer(inputs, initial_state=initial_state)
        if self.output_dropout:
            return self.dropout_layer(inputs)
        else:
            return inputs

    def compute_mask(
        self,
        inputs: tf.Tensor,
        mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        return self.layers[0].compute_mask(inputs, mask)

    def get_config(self):
        return {
            **super().get_config(),
            'num_layers': self.num_layers,
            'units': self.units,
            'projection_size': self.projection_size,
            'recurrent_clip': self.recurrent_clip,
            'projection_clip': self.projection_clip,
            'input_dropout': self.input_dropout,
            'recurrent_dropout': self.recurrent_dropout,
            'output_dropout': self.output_dropout,
            'last_connection': self.last_connection,
            'kernel_initializer': self.kernel_initializer,
            'recurrent_initializer': self.recurrent_initializer,
            'projection_initializer': self.projection_initializer,
            'bias_initializer': self.bias_initializer,
            'kernel_regularizer': self.kernel_regularizer,
            'recurrent_regularizer': self.recurrent_regularizer,
            'projection_regularizer': self.projection_regularizer,
            'bias_regularizer': self.bias_regularizer,
            'kernel_constraint': self.kernel_constraint,
            'recurrent_constraint': self.recurrent_constraint,
            'projection_constraint': self.projection_constraint,
            'bias_constraint': self.bias_constraint
        }



from typing import Optional, Dict, Any, List

import tensorflow as tf
from tensorflow.keras.layers import Embedding

from sknlp.vocab import Vocab
from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint

from .text2vec import Text2vec


class Word2vec(Text2vec):

    def __init__(
        self,
        vocab: Vocab,
        embedding_size: int = 100,
        segmenter: str = 'jieba',
        embeddings_initializer: WeightInitializer = 'uniform',
        embeddings_regularizer: Optional[WeightRegularizer] = None,
        embeddings_constraint: Optional[WeightConstraint] = None,
        name: str = 'word2vec',
        **kwargs
    ) -> None:
        """
        基础符号->向量模块.

        Parameters:
        ----------
        vocab: sknlp.vocab.Vocab. 符号表.
        embed_size: int > 0. 向量维度.
        name: str. 模块名.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
        the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
        the `embeddings` matrix.
        input_length: Length of input sequences, when it is constant.
        This argument is required if you are going to connect
        `Flatten` then `Dense` layers upstream
        (without it, the shape of the dense outputs cannot be computed).

        Inputs
        ----------
        2D tensor with shape: `(batch_size, input_length)`.

        Outputs
        ----------
        3D tensor with shape: `(batch_size, input_length, embed_size)`.
        """
        super().__init__(
            vocab, segmenter=segmenter, name=name, **kwargs
        )
        self._embedding_size = embedding_size
        embedding = Embedding(
            len(vocab),
            embedding_size,
            embeddings_initializer=embeddings_initializer,
            embeddings_regularizer=embeddings_regularizer,
            embeddings_constraint=embeddings_constraint,
            mask_zero=True,
            name='embeddings'
        )
        self._model = tf.keras.Sequential(embedding, name=name)

    def __call__(self, inputs: tf.Tensor) -> tf.Tensor:
        return self._model(inputs)

    @property
    def input_names(self) -> List[str]:
        return ["embeddings_input"]

    @property
    def input_types(self) -> List[str]:
        return ["float"]

    @property
    def input_shapes(self) -> List[List[int]]:
        return [[-1, -1]]

    @property
    def output_names(self) -> List[str]:
        return ["embeddings"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self.embedding_size]]

    @property
    def embedding_size(self):
        return self._embedding_size

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "embedding_size": self._embedding_size
        }

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("max_sequence_length", None)
        return config



from typing import Optional, List, Dict, Any
import os
from collections import Counter

import tensorflow as tf
from official.modeling import activations

from sknlp.vocab import Vocab
from sknlp.layers.bert_layer import BertLayer, BertPreprocessingLayer

from .text2vec import Text2vec


class Bert2vec(Text2vec):

    def __init__(
        self,
        vocab,
        segmenter=None,
        hidden_size=768,
        num_layers=12,
        num_attention_heads=12,
        sequence_length=512,
        max_sequence_length=None,
        type_vocab_size=16,
        intermediate_size=3072,
        activation=activations.gelu,
        dropout_rate=0.1,
        attention_dropout_rate=0.1,
        initializer=tf.keras.initializers.TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs=False,
        name="bert2vec",
        **kwargs
    ):
        super().__init__(vocab, segmenter=segmenter, name=name, **kwargs)
        self._hidden_size = hidden_size
        self._num_layers = num_layers
        self._num_attention_heads = num_attention_heads
        self._sequence_length = sequence_length
        self._max_sequence_length = max_sequence_length
        self._type_vocab_size = type_vocab_size
        self._intermediate_size = intermediate_size
        self._activation = activation
        self._dropout_rate = dropout_rate
        self._attention_dropout_rate = attention_dropout_rate
        self._initializer = initializer
        self._return_all_encoder_outputs = return_all_encoder_outputs

        bert_preprocessing_layer = BertPreprocessingLayer(
            self._vocab.sorted_tokens[2:], sequence_length
        )
        bert_layer = BertLayer(
            len(self._vocab),
            hidden_size=hidden_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            type_vocab_size=type_vocab_size,
            intermediate_size=intermediate_size,
            activation=activation,
            dropout_rate=dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer=initializer,
            return_all_encoder_outputs=return_all_encoder_outputs
        )
        inputs = tf.keras.Input(shape=(1,), dtype=tf.string, name="text_input")
        _, cls = bert_layer(bert_preprocessing_layer(inputs))
        self._model = tf.keras.Model(inputs=inputs, outputs=cls, name=name)

    @property
    def input_names(self) -> List[str]:
        return ["text_input"]

    @property
    def input_types(self) -> List[str]:
        return ["string"]

    @property
    def input_shapes(self) -> List[List[int]]:
        return [[-1, 1]]

    @property
    def output_names(self) -> List[str]:
        return ["transformer_encoder"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self.embedding_size]]

    @property
    def embedding_size(self):
        return self._hidden_size

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "hidden_size": self._hidden_size,
            "num_layers": self._num_layers,
            "num_attention_heads": self._num_attention_heads,
            "sequence_length": self._sequence_length,
            "max_sequence_length": self._max_sequence_length,
            "type_vocab_size": self._type_vocab_size,
            "intermediate_size": self._intermediate_size,
            "attention_dropout_rate": self._attention_dropout_rate,
            "return_all_encoder_outputs": self._return_all_encoder_outputs
        }

    @classmethod
    def from_tfv1_checkpoint(cls, v1_checkpoint, sequence_length=100):
        bert_layer = BertLayer.from_tfv1_checkpoint(v1_checkpoint, sequence_length)
        with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
            vocab_list = f.read().split("\n")
            bert_preprocessing_layer = BertPreprocessingLayer(
                vocab_list[2:], sequence_length
            )
            vocab = Vocab(
                Counter(vocab_list), pad_token=vocab_list[0], unk_token=vocab_list[1]
            )

        module = cls(
            vocab,
            hidden_size=bert_layer.hidden_size,
            num_layers=bert_layer.num_layers,
            num_attention_heads=bert_layer.num_attention_heads,
            sequence_length=bert_layer.sequence_length,
            max_sequence_length=bert_layer.max_sequence_length,
            type_vocab_size=bert_layer.type_vocab_size,
            intermediate_size=bert_layer.intermediate_size,
            activation=bert_layer.activation,
            dropout_rate=bert_layer.dropout_rate,
            attention_dropout_rate=bert_layer.attention_dropout_rate,
            initializer=bert_layer.initializer,
            return_all_encoder_outputs=bert_layer.return_all_encoder_outputs
        )
        inputs = tf.keras.Input(shape=(1,), dtype=tf.string, name="text_input")
        _, cls = bert_layer(bert_preprocessing_layer(inputs))
        module._model = tf.keras.Model(inputs=inputs, outputs=cls, name="bert2vec")
        return module

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("segmenter", None)
        return config

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "BertLayer": BertLayer,
            "BertPreprocessingLayer": BertPreprocessingLayer,
        }



import functools
from tensorflow.keras.metrics import Accuracy
import tensorflow.keras.backend as K


class AccuracyWithLogits(Accuracy):

    def __init__(self,
                 name='acc',
                 dtype=None,
                 logits2scores='sigmoid'):
        super().__init__(name=name,
                         dtype=dtype)
        self.logits2scores = logits2scores
        if logits2scores == 'sigmoid':
            self._l2s = K.sigmoid
        elif logits2scores == 'softmax':
            self._l2s = functools.partial(K.softmax, axis=-1)

    def update_state(self, y_true, y_pred, sample_weight=None):
        super().update_state(y_true, self._l2s(y_pred),
                             sample_weight=sample_weight)

    def get_config(self):
        return {**super().get_config(), 'logits2scores': self.logits2scores}



from typing import Optional, List

import pandas as pd
import tensorflow as tf
from .text_segmenter import get_segmenter


class NLPDataset:

    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        text_segmenter: str = 'char',
        text_dtype: tf.DType = tf.string,
        label_dtype: tf.DType = tf.string,
        text_padding_shape: tuple = (None,),
        label_padding_shape: tuple = (),
        text_padding_value: str = '<pad>',
        label_padding_value: str = ''
    ):
        assert df is not None or csv_file is not None, \
            "either df or csv_file may not be None"
        if df is not None:
            self._dataset = self.dataframe_to_dataset(df)
        elif csv_file is not None:
            self._dataset, self.size = self.load_csv(csv_file, "\t", in_memory)
        self.text_cutter = get_segmenter(text_segmenter)
        self.text_dtype = text_dtype
        self.label_dtype = label_dtype
        self.text_padding_shape = text_padding_shape
        self.label_padding_shape = label_padding_shape
        self.text_padding_value = text_padding_value
        self.label_padding_value = label_padding_value

        self._dataset = self._transform(self._dataset)

    def _text_transform(self, text: tf.Tensor) -> List[str]:
        return self.text_cutter(text.numpy().decode('utf-8'))

    def _label_transform(self, label: tf.Tensor) -> List[str]:
        return label.numpy().decode('utf-8')

    def _transform(self, dataset: tf.data.Dataset) -> tf.data.Dataset:

        def func(text, label):
            return self._text_transform(text), self._label_transform(label)

        return dataset.map(
            lambda t, l: tf.py_function(
                func, inp=[t, l], Tout=(self.text_dtype, self.label_dtype)
            ),
            num_parallel_calls=tf.data.experimental.AUTOTUNE
        )

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: int = 100000
    ) -> tf.data.Dataset:
        dataset = self._dataset
        if shuffle:
            if shuffle_buffer_size <= 0:
                shuffle_buffer_size = 100000
            dataset = dataset.shuffle(shuffle_buffer_size)

        return (
            dataset
            .padded_batch(
                batch_size,
                (self.text_padding_shape, self.label_padding_shape),
                padding_values=(self.text_padding_value,
                                self.label_padding_value)
            )
            .prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
        )

    @classmethod
    def dataframe_to_dataset(cls, df: pd.DataFrame):
        df.fillna('', inplace=True)
        return tf.data.Dataset.from_tensor_slices((df.text, df.label))

    @classmethod
    def load_csv(cls, filename: str, sep: str, in_memory: bool):
        if in_memory:
            df = pd.read_csv(filename, sep=sep, dtype=str)
            return cls.dataframe_to_dataset(df), df.shape[0]
        return (
            tf.data.experimental.CsvDataset(filename,
                                            (tf.string, tf.string),
                                            header=True,
                                            field_delim=sep,
                                            na_value=''),
            -1
        )



import tensorflow as tf


class BertTokenizationLayer(tf.keras.layers.Layer):
    """
    The convention in BERT is:
    (a) For sequence pairs:
    tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
    type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
    (b) For single sequences:
    tokens:   [CLS] the dog is hairy . [SEP]
    type_ids: 0     0   0   0  0     0 0

    Where "type_ids" are used to indicate whether this is the first
    sequence or the second sequence. The embedding vectors for `type=0` and
    `type=1` were learned during pre-training and are added to the wordpiece
    embedding vector (and position vector). This is not *strictly* necessary
    since the [SEP] token unambiguously separates the sequences, but it makes
    it easier for the model to learn the concept of sequences.

    For classification tasks, the first vector (corresponding to [CLS]) is
    used as the "sentence vector". Note that this only makes sense because
    the entire model is fine-tuned.

    The mask has 1 for real tokens and 0 for padding tokens. Only real
    tokens are attended to.
    See more details in https://github.com/tensorflow/models/blob/ad423d065701785587d13d0fe7b566191e7378c6/official/nlp/data/classifier_data_lib.py#L668
    """
    def __init__(self, sep="@!@", name="bert_tokenization", **kwargs):
        self.sep = sep
        super().__init__(name=name, **kwargs)

    def call(self, inputs):

        def padding_string(x):
            return (
                tf.constant(["[CLS]" + self.sep])
                + x
                + tf.constant([self.sep + "[SEP]"])
            )

        padded_inputs = tf.ragged.map_flat_values(padding_string, inputs)
        r = tf.strings.split(padded_inputs, sep=self.sep)
        return tf.concat(
            [
                r[:, :1],
                tf.strings.unicode_split(tf.squeeze(r[:, 1:-1], axis=1), "UTF-8"),
                r[:, -1:]
            ],
            axis=1
        )

    def get_config(self):
        return {**super().get_config(), "sep": self.sep}



from typing import NewType, Callable, Union

import tensorflow as tf


WeightRegularizer = NewType(
    "WeightRegularizer",
    Union[Callable[[tf.Tensor], tf.Tensor], str]
)

WeightInitializer = NewType(
    "WeightInitializer",
    Union[Callable[[tf.Tensor, tf.DType], tf.Tensor], str]
)

WeightConstraint = NewType(
    "WeightConstraint",
    Union[Callable[[tf.Tensor], tf.Tensor], str]
)


__all__ = ["WeightRegularizer", "WeightInitializer", "WeightConstraint"]



from ..text_rnn import TextRNN
from .deep_classifier import DeepClassifier


class TextRNNClassifier(DeepClassifier, TextRNN):

    def __init__(self,
                 classes,
                 is_multilabel=True,
                 segmenter='jieba',
                 max_sequence_length=100,
                 embedding_size=100,
                 num_rnn_layers=1,
                 rnn_hidden_size=512,
                 rnn_projection_size=128,
                 rnn_recurrent_clip=3,
                 rnn_projection_clip=3,
                 rnn_input_dropout=0.5,
                 rnn_recurrent_dropout=0.5,
                 rnn_output_dropout=0.5,
                 num_fc_layers=2,
                 fc_hidden_size=128,
                 rnn_kernel_initializer='glorot_uniform',
                 rnn_recurrent_initializer='orthogonal',
                 rnn_projection_initializer='glorot_uniform',
                 rnn_bias_initializer='zeros',
                 rnn_kernel_regularizer=None,
                 rnn_recurrent_regularizer=None,
                 rnn_projection_regularizer=None,
                 rnn_bias_regularizer=None,
                 rnn_kernel_constraint=None,
                 rnn_recurrent_constraint=None,
                 rnn_projection_constraint=None,
                 rnn_bias_constraint=None,
                 text2vec=None,
                 **kwargs):
        super().__init__(classes,
                         is_multilabel=is_multilabel,
                         segmenter=segmenter,
                         max_sequence_length=max_sequence_length,
                         embedding_size=embedding_size,
                         algorithm='text_rnn',
                         num_rnn_layers=num_rnn_layers,
                         rnn_hidden_size=rnn_hidden_size,
                         rnn_projection_size=rnn_projection_size,
                         rnn_recurrent_clip=rnn_recurrent_clip,
                         rnn_projection_clip=rnn_projection_clip,
                         rnn_input_dropout=rnn_input_dropout,
                         rnn_recurrent_dropout=rnn_recurrent_dropout,
                         rnn_output_dropout=rnn_output_dropout,
                         rnn_last_connection='last',
                         num_fc_layers=num_fc_layers,
                         fc_hidden_size=fc_hidden_size,
                         rnn_kernel_initializer=rnn_kernel_initializer,
                         rnn_recurrent_initializer=rnn_recurrent_initializer,
                         rnn_projection_initializer=rnn_projection_initializer,
                         rnn_bias_initializer=rnn_bias_initializer,
                         rnn_kernel_regularizer=rnn_kernel_regularizer,
                         rnn_recurrent_regularizer=rnn_recurrent_regularizer,
                         rnn_projection_regularizer=rnn_projection_regularizer,
                         rnn_bias_regularizer=rnn_bias_regularizer,
                         rnn_kernel_constraint=rnn_kernel_constraint,
                         rnn_recurrent_constraint=rnn_recurrent_constraint,
                         rnn_projection_constraint=rnn_projection_constraint,
                         rnn_bias_constraint=rnn_bias_constraint,
                         text2vec=text2vec,
                         **kwargs)

    def get_config(self):
        base_config = super().get_config()
        del base_config['rnn_last_connection']
        return base_config



from tensorflow.keras.layers import Dense, Dropout
import tensorflow.keras.backend as K

from ..text_rnn import TextRNN
from .deep_classifier import DeepClassifier


class TextRCNNClassifier(DeepClassifier, TextRNN):

    def __init__(self,
                 classes,
                 is_multilabel=True,
                 segmenter='jieba',
                 max_sequence_length=100,
                 embed_size=100,
                 num_rnn_layers=1,
                 rnn_hidden_size=512,
                 rnn_projection_size=128,
                 rnn_recurrent_clip=3,
                 rnn_projection_clip=3,
                 rnn_input_dropout=0.5,
                 rnn_recurrent_dropout=0.5,
                 rnn_output_dropout=0.5,
                 num_fc_layers=2,
                 fc_hidden_size=128,
                 rnn_kernel_initializer='glorot_uniform',
                 rnn_recurrent_initializer='orthogonal',
                 rnn_projection_initializer='glorot_uniform',
                 rnn_bias_initializer='zeros',
                 rnn_kernel_regularizer=None,
                 rnn_recurrent_regularizer=None,
                 rnn_projection_regularizer=None,
                 rnn_bias_regularizer=None,
                 rnn_kernel_constraint=None,
                 rnn_recurrent_constraint=None,
                 rnn_projection_constraint=None,
                 rnn_bias_constraint=None,
                 text2vec=None,
                 **kwargs):
        super().__init__(classes,
                         is_multilabel=is_multilabel,
                         segmenter=segmenter,
                         max_sequence_length=max_sequence_length,
                         embed_size=embed_size,
                         algorithm='text_rcnn',
                         num_rnn_layers=num_rnn_layers,
                         rnn_hidden_size=rnn_hidden_size,
                         rnn_projection_size=rnn_projection_size,
                         rnn_recurrent_clip=rnn_recurrent_clip,
                         rnn_projection_clip=rnn_projection_clip,
                         rnn_input_dropout=rnn_input_dropout,
                         rnn_recurrent_dropout=rnn_recurrent_dropout,
                         rnn_output_dropout=rnn_output_dropout,
                         rnn_last_connection='all',
                         num_fc_layers=num_fc_layers,
                         fc_hidden_size=fc_hidden_size,
                         rnn_kernel_initializer=rnn_kernel_initializer,
                         rnn_recurrent_initializer=rnn_recurrent_initializer,
                         rnn_projection_initializer=rnn_projection_initializer,
                         rnn_bias_initializer=rnn_bias_initializer,
                         rnn_kernel_regularizer=rnn_kernel_regularizer,
                         rnn_recurrent_regularizer=rnn_recurrent_regularizer,
                         rnn_projection_regularizer=rnn_projection_regularizer,
                         rnn_bias_regularizer=rnn_bias_regularizer,
                         rnn_kernel_constraint=rnn_kernel_constraint,
                         rnn_recurrent_constraint=rnn_recurrent_constraint,
                         rnn_projection_constraint=rnn_projection_constraint,
                         rnn_bias_constraint=rnn_bias_constraint,
                         text2vec=text2vec,
                         **kwargs)

    def build_encode_layer(self, inputs):
        rnn_outputs = super().build_encode_layer(inputs)
        if self._rnn_input_dropout:
            noise_shape = (None, 1, self._embed_size)
            inputs = Dropout(self._rnn_input_dropout,
                             noise_shape=noise_shape,
                             name='embed_dropout')(inputs)
        mixed_inputs = K.concatenate([inputs, rnn_outputs], axis=-1)
        mixed_outputs = Dense(self._fc_hidden_size,
                              activation='tanh')(mixed_inputs)
        return K.max(mixed_outputs, axis=1)

    def get_config(self):
        base_config = super().get_config()
        del base_config['rnn_last_connection']
        return base_config



from .text_rnn import TextRNNClassifier
from .text_rcnn import TextRCNNClassifier
from .bert_classifier import BertClassifier


__all__ = ['TextRNNClassifier', 'TextRCNNClassifier', 'BertClassifier']



from typing import Sequence, Optional, Dict, Any, Callable, List

import json
import os
import itertools
from collections import Counter

import tensorflow as tf

from ..vocab import Vocab


class BaseNLPModel:

    def __init__(
        self,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        name: Optional[str] = None
    ) -> None:
        self._max_sequence_length = max_sequence_length
        self._sequence_length = sequence_length
        self._name = name
        self._built = False

    @staticmethod
    def build_vocab(
        texts: Sequence[str],
        segment_func: Callable[[str], Sequence[str]],
        min_frequency=5
    ) -> Vocab:
        counter = Counter(
            itertools.chain.from_iterable(segment_func(text) for text in texts)
        )
        return Vocab(counter, min_frequency=min_frequency)

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def build(self) -> None:
        self._model = tf.keras.Model(
            inputs=self.get_inputs(), outputs=self.get_outputs(), name=self._name
        )
        self._built = True

    def get_inputs(self) -> tf.Tensor:
        raise NotImplementedError()

    @property
    def input_names(self) -> List[str]:
        return []

    @property
    def input_types(self) -> List[str]:
        return []

    @property
    def input_shapes(self) -> List[List[int]]:
        return []

    def get_outputs(self) -> tf.Tensor:
        raise NotImplementedError()

    @property
    def output_names(self) -> List[str]:
        return []

    @property
    def output_types(self) -> List[str]:
        return []

    @property
    def output_shapes(self) -> List[List[int]]:
        return []

    def get_loss(self):
        raise NotImplementedError()

    def get_metrics(self):
        raise NotImplementedError()

    def get_callbacks(self):
        raise NotImplementedError()

    def get_custom_objects(self):
        return {}

    def freeze(self) -> None:
        for layer in self._model.layers:
            layer.trainable = False

    def save_config(self, directory: str, filename: str = "meta.json") -> None:
        with open(os.path.join(directory, filename), "w") as f:
            f.write(json.dumps(self.get_config()))

    def save(self, directory: str) -> None:
        self._model.save(os.path.join(directory, "model"), save_format="tf")
        self.save_config(directory)

    @classmethod
    def load(cls, directory: str) -> "BaseNLPModel":
        with open(os.path.join(directory, "meta.json")) as f:
            meta = json.loads(f.read())
        module = cls.from_config(meta)
        module._model = tf.keras.models.load_model(
            os.path.join(directory, "model"),
            custom_objects=module.get_custom_objects()
        )
        module._built = True
        return module

    def export(self, directory: str, name: str, version: str = "0") -> None:
        d = os.path.join(directory, name, version)
        self._model.save(d, save_format="tf")
        self.save_config(d)

    def get_config(self) -> Dict[str, Any]:
        return {
            "inputs": self.input_names,
            "input_types": self.input_types,
            "input_shapes": self.input_shapes,
            "outputs": self.output_names,
            "output_types": self.output_types,
            "output_shapes": self.output_shapes,
            "max_sequence_length": self._max_sequence_length,
            "sequence_length": self._sequence_length,
            "name": self._name
        }

    @classmethod
    def _filter_config(cls, config):
        config.pop("inputs", None)
        config.pop("input_types", None)
        config.pop("input_shapes", None)
        config.pop("outputs", None)
        config.pop("output_types", None)
        config.pop("output_shapes", None)
        return config

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "BaseNLPModel":
        return cls(**cls._filter_config(config))



from .precision_recall_f1 import PrecisionWithLogits, RecallWithLogits
from .utils import logits2scores, logits2classes, scores2classes


__all__ = ['PrecisionWithLogits', 'RecallWithLogits',
           'logits2scores', 'logits2classes', 'scores2classes']



from typing import Optional, List, Union, Sequence, Dict

import json
from collections import defaultdict, Counter


class Vocab:

    def __init__(
        self,
        counter: Optional[Counter] = None,
        min_frequency: int = 1,
        pad_token: str = "<pad>",
        unk_token: str = "<unk>"
    ) -> None:
        self._min_frequency = min_frequency
        self._unk_token = unk_token
        self._pad_token = pad_token

        self._token2idx = defaultdict(lambda: 1)
        self._token2idx[pad_token] = 0
        self._token2idx[unk_token] = 1
        self._reversed_tokens = {unk_token, pad_token}

        self._token_frequency = defaultdict(lambda: 0)
        if counter is not None:
            for _, (token, frequency) in enumerate(counter.most_common()):
                if token not in self._reversed_tokens and frequency >= min_frequency:
                    self._token2idx[token] = len(self._token2idx)
                    self._token_frequency[token] = frequency
        self._idx2token = dict(zip(self._token2idx.values(), self._token2idx.keys()))

    def set_vocab(
        self,
        token2idx: Dict[str, int],
        pad_token: str,
        unk_token: str
    ) -> None:
        if token2idx.get(pad_token, None) != 0:
            raise ValueError("padding token should have index 0")
        if token2idx.get(unk_token, None) != 1:
            raise ValueError("unknown token should have index 1")
        self._token2idx = defaultdict(lambda: 1)
        self._token2idx.update(**token2idx)
        self._idx2token = dict(zip(self._token2idx.values(), self._token2idx.keys()))

    def idx2token(self, indices: Union[int, Sequence[int]]) -> Union[str, List[str]]:
        """
        Lookup tokens by indices.

        raise KeyError if any indices are greater than or equal to
        size of vacab.

        Parameters
        ----------
        indices: int or list of int

        Returns
        ----------
        str or list of str
        """
        if isinstance(indices, int):
            idx = indices
            if idx not in self._idx2token:
                raise KeyError('index %d is out of vacab' % idx)
            return self._idx2token[idx]
        elif isinstance(indices, (list, tuple)):
            res = []
            for idx in indices:
                if idx not in self._idx2token:
                    raise KeyError('index %d is out of vacab' % idx)
                res.append(self.idx2token(idx))
            return res
        else:
            raise ValueError(
                "indices should be int or a list of int, "
                "but %s was given" % type(indices)
            )

    def token2idx(self, tokens: Union[str, List[str]]) -> Union[int, List[int]]:
        if isinstance(tokens, str):
            return self._token2idx[tokens]
        elif isinstance(tokens, (list, tuple)):
            return [self.token2idx(token) for token in tokens]
        else:
            raise ValueError(
                "tokens should be str or a list of str, "
                "but %s was given" % type(tokens)
            )

    def to_json(self) -> str:
        return json.dumps(
            {
                'tokens': self._token_frequency,
                'token2idx': self._token2idx,
                'unk': self.unk,
                'pad': self.pad
            },
            ensure_ascii=False
        )

    @classmethod
    def from_json(cls, json_str: str) -> "Vocab":
        vocab_dict = json.loads(json_str)
        vocab = cls(
            min_frequency=1,
            pad_token=vocab_dict['pad'],
            unk_token=vocab_dict['unk']
        )
        vocab.set_vocab(vocab_dict["token2idx"], vocab_dict['pad'], vocab_dict['unk'])
        return vocab

    def __getitem__(self, tokens: Union[str, Sequence[str]]) -> List[int]:
        return self.token2idx(tokens)

    def __contains__(self, token: str) -> bool:
        return token in self._token2idx

    def __len__(self) -> int:
        return len(self._token2idx)

    def __repr__(self) -> str:
        return 'Vocab(size=%d, unk="%s")' % (len(self), self._unk_token)

    @property
    def pad(self) -> str:
        return self._pad_token

    @property
    def unk(self) -> str:
        return self._unk_token

    @property
    def sorted_tokens(self) -> List[str]:
        items = sorted(self._token2idx.items(), key=lambda x: x[1])
        return [k for k, v in items]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sknlp',
 'sknlp.callbacks',
 'sknlp.data',
 'sknlp.data.text_segmenter',
 'sknlp.layers',
 'sknlp.metrics',
 'sknlp.module',
 'sknlp.module.classifiers',
 'sknlp.module.text2vec',
 'sknlp.typing',
 'sknlp.utils',
 'sknlp.vocab']

package_data = \
{'': ['*']}

install_requires = \
['jieba_fast>=0.53,<0.54',
 'joblib>=0.14.1,<0.15.0',
 'pandas==1.0.4',
 'sklearn>=0.0,<0.1',
 'tabulate>=0.8.6,<0.9.0',
 'tensorflow-text>=2.2.0,<3.0.0',
 'tensorflow==2.2.0',
 'tf-models-official>=2.2.0,<3.0.0']

setup_kwargs = {
    'name': 'sknlp',
    'version': '0.1.5',
    'description': '',
    'long_description': None,
    'author': 'nanaya',
    'author_email': 'nanaya100@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import tensorflow as tf


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


def f_score(p, r, beta):
    return (1 + beta ** 2) * (p * r) / (beta ** 2 * p + r) if p + r != 0 else 0


class FScore(tf.keras.callbacks.Callback):

    def __init__(self, beta=1):
        super().__init__()
        self.beta = beta

    def on_epoch_end(self, epoch, logs=None):
        p, r = logs['precision'], logs['recall']
        if 'val_precision' in logs and 'val_recall' in logs:
            val_p, val_r = logs['val_precision'], logs['val_recall']
            logs['val_f-score'] = f_score(val_p, val_r, self.beta)
        else:
            logs['val_f-score'] = 0
        logs['f-score'] = f_score(p, r, self.beta)
        logger.info('f-score: %.2f, val_f-score: %.2f' % (
            logs['f-score'] * 100, logs['val_f-score'] * 100
        ))



from typing import Sequence, Union, Optional, Dict, Any, List

import os
import logging

import tensorflow as tf
import pandas as pd

from ..data import NLPDataset
from ..data.text_segmenter import get_segmenter
from ..vocab import Vocab
from .text2vec import Word2vec, Text2vec
from .base_model import BaseNLPModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)



class SupervisedNLPModel(BaseNLPModel):

    def __init__(
        self,
        classes: Sequence[str],
        max_sequence_length: int = 100,
        sequence_length: Optional[int] = None,
        segmenter: str = "jieba",
        embedding_size: int = 100,
        text2vec: Optional[Text2vec] = None,
        task: Optional[str] = None,
        algorithm: Optional[str] = None,
        **kwargs
    ):
        super().__init__(
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            **kwargs
        )
        self._text2vec = text2vec
        if text2vec is not None:
            self._segmenter = text2vec.segmenter
            self._embedding_size = text2vec.embedding_size
            self._max_sequence_length = (
                text2vec._max_sequence_length or max_sequence_length
            )
            self._sequence_length = text2vec._sequence_length
        else:
            self._segmenter = segmenter
            self._embedding_size = embedding_size
            self._max_sequence_length = max_sequence_length
            self._sequence_length = sequence_length

        self._task = task
        self._algorithm = algorithm
        self._class2idx = dict(zip(list(classes), range(len(classes))))
        self._idx2class = dict(zip(range(len(classes)), list(classes)))
        self._num_classes = len(self._class2idx)

    def get_inputs(self) -> tf.Tensor:
        return self._text2vec.get_inputs()

    @property
    def input_names(self) -> List[str]:
        return self._text2vec.input_names if self._text2vec else []

    @property
    def input_types(self) -> List[str]:
        return self._text2vec.input_types if self._text2vec else []

    @property
    def input_shapes(self) -> List[List[int]]:
        return self._text2vec.input_shapes if self._text2vec else []

    def get_outputs(self) -> tf.Tensor:
        inputs = self._text2vec.get_outputs()
        return self.build_output_layer(self.build_encode_layer(inputs))

    @classmethod
    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str]
    ) -> NLPDataset:
        raise NotImplementedError()

    def prepare_tf_dataset(
        self,
        X: Sequence[str],
        y: Sequence[str],
        dataset: NLPDataset,
        batch_size: int,
        shuffle: bool = True
    ) -> tf.data.Dataset:
        assert not ((X is None or y is None) and dataset is None)

        if self._text2vec is None:
            cut = get_segmenter(self._segmenter)
            vocab = self.build_vocab(X, cut)
            self._text2vec = Word2vec(vocab, self._embedding_size, self._segmenter)

        if dataset is not None:
            return dataset.batchify(batch_size, shuffle=shuffle)

        if isinstance(y[0], (list, tuple)):
            y = ["|".join(map(str, y_i)) for y_i in y]
        df = pd.DataFrame(zip(X, y), columns=["text", "label"])
        dataset = self.create_dataset_from_df(
            df, self._text2vec.vocab, self._text2vec.segmenter, self._class2idx.keys()
        )
        return dataset.batchify(
            batch_size, shuffle=shuffle, shuffle_buffer_size=df.shape[0]
        )

    def fit(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        valid_X: Sequence[str] = None,
        valid_y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        valid_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 30,
        optimizer: str = "adam",
        lr: float = 1e-3,
        lr_update_factor: float = 0.5,
        lr_update_epochs: int = 10,
        clip: float = 5.0,
        checkpoint: Optional[str] = None,
        save_frequency: int = 1,
        log_file: Optional[str] = None,
        verbose: int = 2
    ) -> None:
        assert not (self._text2vec is None and X is None), \
            "When token2vec is not given, X must be provided"

        self.train_dataset = self.prepare_tf_dataset(X, y, dataset, batch_size)
        if ((valid_X is None or valid_y is None) and valid_dataset is None):
            self.valid_dataset = None
        else:
            self.valid_dataset = self.prepare_tf_dataset(
                valid_X, valid_y, valid_dataset, batch_size, shuffle=False
            )
        if not self._built:
            self.build()
        optimizer = tf.keras.optimizers.Adam(learning_rate=lr, clipnorm=clip)
        if optimizer != "adam":
            pass
        loss = self.get_loss()
        metrics = self.get_metrics()
        callbacks = self.get_callbacks()
        self._model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        def scheduler(epoch):
            new_lr = lr * lr_update_factor ** (epoch // lr_update_epochs)
            if epoch != 0 and epoch % lr_update_epochs == 0:
                logger.info("change learning rate to %.4g" % new_lr)
            return new_lr

        lr_decay = tf.keras.callbacks.LearningRateScheduler(scheduler)
        callbacks = [*callbacks, lr_decay]
        if log_file is not None:
            callbacks.append(tf.keras.callbacks.CSVLogger(log_file))
        self._model.fit(
            self.train_dataset,
            epochs=n_epochs,
            validation_data=self.valid_dataset,
            callbacks=callbacks,
            verbose=verbose
        )

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: NLPDataset = None,
        batch_size: int = 128
    ) -> tf.Tensor:
        assert self._built
        y = None if X is None else ["O" for _ in range(len(X))]
        dataset = self.prepare_tf_dataset(X, y, dataset, batch_size, shuffle=False)

        def _f(x, y):
            return x

        return self._model.predict(dataset.map(_f))

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        thresholds: Union[float, Dict[str, float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "classes": list(self._class2idx.keys()),
            "task": self._task,
            "algorithm": self._algorithm
        }

    def save(self, directory):
        super().save(directory)
        self._text2vec.save_vocab(directory)

    @classmethod
    def load(cls, directory):
        module = super().load(directory)
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        module._text2vec = Text2vec(vocab)
        return module

    def export(self, directory, name, version="0"):
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self._text2vec.save_vocab(d)



import os
import tempfile
import logging

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import InputSpec
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from official.modeling import activations
from official.nlp.modeling import layers
from official.nlp.modeling import networks
from official.nlp.bert import configs
import tensorflow.compat.v1 as tfv1

from .bert_tokenization import BertTokenizationLayer


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


class BertPreprocessingLayer(tf.keras.layers.Layer):

    def __init__(self, vocab, sequence_length=100, **kwargs):
        self.vocab = vocab
        self.sequence_length = sequence_length
        self.text2vec = TextVectorization(
            max_tokens=len(vocab) + 3,
            output_sequence_length=sequence_length,
            standardize=None,
            split=BertTokenizationLayer()
        )
        self.text2vec.set_vocabulary(vocab)
        super().__init__(**kwargs)

    def call(self, inputs):
        return self.text2vec(inputs)

    def get_config(self):
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "sequence_length": self.sequence_length
        }

BERT_NAME_REPLACEMENTS = (
    ("bert/", ""),
    ("encoder", "transformer"),
    ("embeddings/word_embeddings", "word_embeddings/embeddings"),
    ("embeddings/token_type_embeddings", "type_embeddings/embeddings"),
    ("embeddings/position_embeddings", "position_embedding/embeddings"),
    ("embeddings/LayerNorm", "embeddings/layer_norm"),
    ("attention/self", "self_attention"),
    ("attention/output/dense", "self_attention_output"),
    ("attention/output/LayerNorm", "self_attention_layer_norm"),
    ("intermediate/dense", "intermediate"),
    ("output/dense", "output"),
    ("output/LayerNorm", "output_layer_norm"),
    ("pooler/dense", "pooler_transform"),
    ("cls/predictions/output_bias", "cls/predictions/output_bias/bias"),
    ("cls/seq_relationship/output_bias", "predictions/transform/logits/bias"),
    ("cls/seq_relationship/output_weights", "predictions/transform/logits/kernel"),
)
BERT_PERMUTATIONS = (("cls/seq_relationship/output_weights", (1, 0)),)


def _bert_name_replacement(var_name, name_replacements):
  """Gets the variable name replacement."""
  for src_pattern, tgt_pattern in name_replacements:
    if src_pattern in var_name:
      old_var_name = var_name
      var_name = var_name.replace(src_pattern, tgt_pattern)
      logger.info("Converted: %s --> %s", old_var_name, var_name)
  return var_name


def _has_exclude_patterns(name, exclude_patterns):
  """Checks if a string contains substrings that match patterns to exclude."""
  for p in exclude_patterns:
    if p in name:
      return True
  return False


def _get_permutation(name, permutations):
  """Checks whether a variable requires transposition by pattern matching."""
  for src_pattern, permutation in permutations:
    if src_pattern in name:
      logger.info("Permuted: %s --> %s", name, permutation)
      return permutation

  return None


def _get_new_shape(name, shape, num_heads):
  """Checks whether a variable requires reshape by pattern matching."""
  if "self_attention_output/kernel" in name:
    return (num_heads, shape[0] // num_heads, shape[1])
  if "self_attention_output/bias" in name:
    return shape

  patterns = [
      "self_attention/query", "self_attention/value", "self_attention/key"
  ]
  for pattern in patterns:
    if pattern in name:
      if "kernel" in name:
        return tuple([shape[0], num_heads, shape[1] // num_heads])
      if "bias" in name:
        return tuple([num_heads, shape[0] // num_heads])
  return None


def convert_checkpoint(checkpoint_from_path,
                       checkpoint_to_path,
                       num_heads,
                       name_replacements,
                       permutations,
                       exclude_patterns=None):
  """Migrates the names of variables within a checkpoint.
  Args:
    checkpoint_from_path: Path to source checkpoint to be read in.
    checkpoint_to_path: Path to checkpoint to be written out.
    num_heads: The number of heads of the model.
    name_replacements: A list of tuples of the form (match_str, replace_str)
      describing variable names to adjust.
    permutations: A list of tuples of the form (match_str, permutation)
      describing permutations to apply to given variables. Note that match_str
      should match the original variable name, not the replaced one.
    exclude_patterns: A list of string patterns to exclude variables from
      checkpoint conversion.
  Returns:
    A dictionary that maps the new variable names to the Variable objects.
    A dictionary that maps the old variable names to the new variable names.
  """
  with tfv1.Graph().as_default():
    logger.info("Reading checkpoint_from_path %s", checkpoint_from_path)
    reader = tfv1.train.load_checkpoint(checkpoint_from_path)
    name_shape_map = reader.get_variable_to_shape_map()
    new_variable_map = {}
    conversion_map = {}
    for var_name in name_shape_map:
      if exclude_patterns and _has_exclude_patterns(var_name, exclude_patterns):
        continue
      # Get the original tensor data.
      tensor = reader.get_tensor(var_name)

      # Look up the new variable name, if any.
      new_var_name = _bert_name_replacement(var_name, name_replacements)

      # See if we need to reshape the underlying tensor.
      new_shape = None
      if num_heads > 0:
        new_shape = _get_new_shape(new_var_name, tensor.shape, num_heads)
      if new_shape:
        logger.info("Veriable %s has a shape change from %s to %s",
                    var_name, tensor.shape, new_shape)
        tensor = np.reshape(tensor, new_shape)

      # See if we need to permute the underlying tensor.
      permutation = _get_permutation(var_name, permutations)
      if permutation:
        tensor = np.transpose(tensor, permutation)

      # Create a new variable with the possibly-reshaped or transposed tensor.
      var = tfv1.Variable(tensor, name=var_name)

      # Save the variable into the new variable map.
      new_variable_map[new_var_name] = var

      # Keep a list of converter variables for sanity checking.
      if new_var_name != var_name:
        conversion_map[var_name] = new_var_name

    saver = tfv1.train.Saver(new_variable_map)

    with tfv1.Session() as sess:
      sess.run(tfv1.global_variables_initializer())
      logger.info("Writing checkpoint_to_path %s", checkpoint_to_path)
      saver.save(sess, checkpoint_to_path, write_meta_graph=True)

  logger.info("Summary:")
  logger.info("  Converted %d variable name(s).", len(new_variable_map))
  logger.info("  Converted: %s", str(conversion_map))



def create_bert_model(cfg):
    """Creates a BERT keras core model from BERT configuration.
    Args:
    cfg: A `BertConfig` to create the core model.
    Returns:
    A TransformerEncoder netowork.
    """
    bert_encoder = networks.TransformerEncoder(
      vocab_size=cfg.vocab_size,
      hidden_size=cfg.hidden_size,
      num_layers=cfg.num_hidden_layers,
      num_attention_heads=cfg.num_attention_heads,
      intermediate_size=cfg.intermediate_size,
      activation=activations.gelu,
      dropout_rate=cfg.hidden_dropout_prob,
      attention_dropout_rate=cfg.attention_probs_dropout_prob,
      sequence_length=80,
      max_sequence_length=cfg.max_position_embeddings,
      type_vocab_size=cfg.type_vocab_size,
      initializer=tf.keras.initializers.TruncatedNormal(stddev=cfg.initializer_range)
    )
    return bert_encoder


class BertLayer(tf.keras.layers.Layer):
    """
    copy from https://github.com/tensorflow/models/blob/v2.2.0/official/nlp/modeling/networks/transformer_encoder.py.
    """

    def __init__(
        self,
        vocab_size,
        hidden_size=768,
        num_layers=12,
        num_attention_heads=12,
        sequence_length=512,
        max_sequence_length=None,
        type_vocab_size=16,
        intermediate_size=3072,
        activation=activations.gelu,
        dropout_rate=0.1,
        attention_dropout_rate=0.1,
        initializer=None, # tf.keras.initializers.TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs=False,
        name="bert_layer",
        **kwargs
    ):
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.sequence_length = sequence_length
        self.max_sequence_length = max_sequence_length
        self.type_vocab_size = type_vocab_size
        self.intermediate_size = intermediate_size
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.attention_dropout_rate = attention_dropout_rate
        self.initializer = initializer
        self.return_all_encoder_outputs = return_all_encoder_outputs
        self.input_spec = InputSpec(ndim=2, shape=(None, self.sequence_length))

        self.embedding_layer = layers.OnDeviceEmbedding(
            vocab_size=vocab_size,
            embedding_width=hidden_size,
            initializer=initializer,
            name="word_embeddings"
        )
        # Always uses dynamic slicing for simplicity.
        self.position_embedding_layer = layers.PositionEmbedding(
            initializer=initializer,
            use_dynamic_slicing=True,
            max_sequence_length=max_sequence_length,
            name="position_embeddings"
        )
        self.type_embedding_layer = layers.OnDeviceEmbedding(
            vocab_size=type_vocab_size,
            embedding_width=hidden_size,
            initializer=initializer,
            use_one_hot=True,
            name="type_embeddings"
        )
        self.normalize_layer = tf.keras.layers.LayerNormalization(
            name="embeddings/layer_norm",
            axis=-1,
            epsilon=1e-12,
            dtype=tf.float32
        )
        self.transformer_layers = []
        for i in range(num_layers):
            layer = layers.Transformer(
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                intermediate_activation=activation,
                dropout_rate=dropout_rate,
                attention_dropout_rate=attention_dropout_rate,
                kernel_initializer=initializer,
                name="transformer/layer_%d" % i
            )
            self.transformer_layers.append(layer)

        self.cls_output_layer = tf.keras.layers.Dense(
            units=hidden_size,
            activation="tanh",
            kernel_initializer=initializer,
            name="pooler_transform"
        )

        super().__init__(name=name, **kwargs)

    def call(self, inputs):
        input_masks = inputs != 0
        input_types = tf.zeros_like(inputs, name="input_type_ids")

        word_embeddings = self.embedding_layer(inputs)
        position_embeddings = self.position_embedding_layer(word_embeddings)
        type_embeddings = self.type_embedding_layer(input_types)
        embeddings = word_embeddings + position_embeddings + type_embeddings
        embeddings = self.normalize_layer(embeddings)
        embeddings = tf.keras.layers.Dropout(rate=self.dropout_rate)(embeddings)

        data = embeddings
        attention_mask = layers.SelfAttentionMask()([data, input_masks])
        encoder_outputs = []
        for layer in self.transformer_layers:
            data = layer([data, attention_mask])
            encoder_outputs.append(data)

        first_token_tensor = tf.keras.layers.Lambda(
            lambda x: tf.squeeze(x[:, 0:1, :], axis=1)
        )(encoder_outputs[-1])
        cls_output = self.cls_output_layer(first_token_tensor)
        if self.return_all_encoder_outputs:
            return [encoder_outputs, cls_output]
        else:
            return [encoder_outputs[-1], cls_output]

    def get_config(self):
        return {
            **super().get_config(),
            "vocab_size": self.vocab_size,
            "hidden_size": self.hidden_size,
            "num_layers": self.num_layers,
            "num_attention_heads": self.num_attention_heads,
            "sequence_length": self.sequence_length,
            "max_sequence_length": self.max_sequence_length,
            "type_vocab_size": self.type_vocab_size,
            "intermediate_size": self.intermediate_size,
            "activation": tf.keras.activations.serialize(self.activation),
            "dropout_rate": self.dropout_rate,
            "attention_dropout_rate": self.attention_dropout_rate,
            "initializer": self.initializer,
            "return_all_encoder_outputs": self.return_all_encoder_outputs
        }

    @classmethod
    def from_tfv1_checkpoint(cls, v1_checkpoint, sequence_length=100):
        bert_config = configs.BertConfig.from_json_file(
            os.path.join(v1_checkpoint, "bert_config.json")
        )
        layer = cls(
            vocab_size=bert_config.vocab_size,
            hidden_size=bert_config.hidden_size,
            num_layers=bert_config.num_hidden_layers,
            num_attention_heads=bert_config.num_attention_heads,
            intermediate_size=bert_config.intermediate_size,
            activation=activations.gelu,
            dropout_rate=bert_config.hidden_dropout_prob,
            attention_dropout_rate=bert_config.attention_probs_dropout_prob,
            sequence_length=sequence_length,
            max_sequence_length=bert_config.max_position_embeddings,
            type_vocab_size=bert_config.type_vocab_size,
            initializer=tf.keras.initializers.TruncatedNormal(
                stddev=bert_config.initializer_range
            )
        )
        bert_network = create_bert_model(bert_config)
        with tempfile.TemporaryDirectory() as temp_dir:
            temporary_checkpoint = os.path.join(temp_dir, "ckpt")
            convert_checkpoint(
                checkpoint_from_path=v1_checkpoint,
                checkpoint_to_path=temporary_checkpoint,
                num_heads=bert_config.num_attention_heads,
                name_replacements=BERT_NAME_REPLACEMENTS,
                permutations=BERT_PERMUTATIONS,
                exclude_patterns=["adam", "Adam"]
            )
            bert_network.load_weights(
                temporary_checkpoint
            ).assert_existing_objects_matched()

            with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
                vocab = f.read().split("\n")
                bert_preprocessing_layer = BertPreprocessingLayer(
                    vocab[2:], sequence_length
                )

            layer(bert_preprocessing_layer(tf.constant([["是的"], ["不是的"]])))
            layer.embedding_layer.set_weights(
                bert_network.get_layer("word_embeddings").get_weights()
            )
            layer.position_embedding_layer.set_weights(
                bert_network.get_layer("position_embedding").get_weights()
            )
            layer.type_embedding_layer.set_weights(
                bert_network.get_layer("type_embeddings").get_weights()
            )
            layer.normalize_layer.set_weights(
                bert_network.get_layer("embeddings/layer_norm").get_weights()
            )
            for i in range(bert_config.num_hidden_layers):
                layer.transformer_layers[i].set_weights(
                    bert_network.get_layer("transformer/layer_%d" % i).get_weights()
                )
            layer.cls_output_layer.set_weights(
                bert_network.get_layer("pooler_transform").get_weights()
            )
            return layer



# -*- coding: utf-8 -*-
from .classification_dataset import ClassificationDataset
from .nlp_dataset import NLPDataset
from .bert_dataset import BertClassificationDataset


__all__ = [
    'NLPDataset', 'ClassificationDataset', 'BertClassificationDataset'
]



from typing import List, Tuple, Union, Optional, Sequence

import numpy as np
import pandas as pd

from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.metrics import multilabel_confusion_matrix
from scipy.special import expit, softmax



def logits2probabilities(logits: np.ndarray, is_multilabel: bool) -> np.ndarray:
    """
    计算logits的sigmoid或softmax分值.

    Parameters
    ----------
    logits: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类, 多标签分值使用sigmoid, 非多标签使用softmax

    Returns
    ----------
    返回一个shape(`n_samples`, `n_classes`)的``ndarray``.
    """
    if logits.ndim != 2:
        raise ValueError(
            "logits should have shape(`n_samples`, `n_classes`),"
            "but shape%s was given" % str(logits.shape)
        )
    if is_multilabel:
        return expit(logits)
    else:
        return softmax(logits, axis=1)


def _validate_thresholds(
    thresholds: Union[Sequence[float], float],
    length: int
) -> List[float]:
    if isinstance(thresholds, float):
        return [thresholds for _ in range(length)]
    elif len(thresholds) != length:
        raise ValueError(
            "thresholds should have length %d, "
            "but length %d was given" % (length, len(thresholds))
        )
    return list(thresholds)


def probabilities2classes(
    probabilities: np.ndarray,
    is_multilabel: bool,
    thresholds: Union[Sequence[float], float] = 0.5
) -> Union[List[int], List[List[int]]]:
    """
    根据给定的`threshold`, 将分类问题的`scores`解析为对应的类别.

    Parameters
    ----------
    scores: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类
    threshold: 正例判断的阈值(仅在多标签时有作用)

    Returns
    ----------
    返回一个长度为`n_samples`的``list``.

    如果是多标签分类, 每一个sample对应的结果为一个``list``,
    其中的每个``int``值为这个sample对应的类别.

    如果不是多标签分类, 则每一个sample对应的结果为一个``int``,
    表示这个sample对应的类别.
    """
    if probabilities.ndim != 2:
        raise ValueError(
            "probabilities should have shape(`n_samples`, `n_classes`), "
            "but shape%s was given" % str(probabilities.shape)
        )
    if is_multilabel:
        thresholds = _validate_thresholds(thresholds, probabilities.shape[1])
        return [
            np.where(probabilities[i, :] > thresholds)[0].tolist()
            for i in range(probabilities.shape[0])
        ]
    else:
        return np.argmax(probabilities, axis=1).tolist()


def logits2classes(
    logits: np.ndarray,
    is_multilabel: bool,
    thresholds: Union[Sequence[float], float] = 0.5
) -> Union[List[int], List[List[int]]]:
    """
    根据给定的`threshold`, 将分类问题的`logits`解析为对应的类别.

    Parameters
    ----------
    logits: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类
    threshold: 正例判断的阈值(仅在多标签时有作用)

    Returns
    ----------
    返回一个长度为`n_samples`的``list``.

    如果是多标签分类, 每一个sample对应的结果为一个``list``,
    其中的每个``int``值为这个sample对应的类别.

    如果不是多标签分类, 则每一个sample对应的结果为一个``int``,
    表示这个sample对应的类别.
    """
    if logits.ndim != 2:
        raise ValueError(
            "logits should have shape(`n_samples`, `n_classes`),"
            "but shape%s was given" % str(logits.shape)
        )
    if is_multilabel:
        thresholds = _validate_thresholds(thresholds, logits.shape[1])
        return [
            np.where(expit(logits[i, :]) > thresholds)[0].tolist()
            for i in range(logits.shape[0])
        ]
    else:
        return np.argmax(logits, axis=1).tolist()


def precision_recall_fscore(
    tp: int, fp: int, fn: int, beta: float = 1
) -> Tuple[float, float, float]:
    precision = tp / (tp + fp) if tp + fp != 0 else 0
    recall = tp / (tp + fn) if tp + fn != 0 else 0
    if precision + recall != 0:
        fscore = (1 + beta ** 2) * precision * recall / (beta ** 2 * precision + recall)
    else:
        fscore = 0
    return precision, recall, fscore


def label_binarizer(
    y: Union[Sequence[Sequence[str]], Sequence[str]],
    p: Union[Sequence[Sequence[str]], Sequence[str]],
    classes: Optional[Sequence[str]] = None,
) -> Tuple[np.ndarray, np.ndarray, List[str]]:
    if isinstance(y[0], str):
        y = [[yi] for yi in y]
    if isinstance(p[0], str):
        p = [[pi] for pi in p]
    binarizer = MultiLabelBinarizer(classes=classes)
    return (
        binarizer.fit_transform(y),
        binarizer.transform(p),
        binarizer.classes_.tolist()
    )


def classification_fscore(
    y: Union[Sequence[Sequence[str]], Sequence[str]],
    p: Union[Sequence[Sequence[str]], Sequence[str]],
    is_multilabel: bool,
    classes: Optional[Sequence[str]] = None,
    beta: float = 1
) -> pd.DataFrame:
    """
    计算分类结果的F值.

    Parameters
    ----------
    y: 标注标签
    p: 预测标签
    classes: 所有的标签集合, 如果不提供则取y
    is_multilabel: 是否是多标签分类
    beta: 加权值, 默认为1, f = (1 + beta**2) * (P * R) / (beta**2 * P + R)

    Returns
    ----------
    返回一个pd.DataFrame.

    """
    y_one_hot, p_one_hot, classes = label_binarizer(y, p, classes)
    records = []
    for class_, arr in zip(
        classes, multilabel_confusion_matrix(y_one_hot, p_one_hot)
    ):
        tp, fp, fn, tn = arr[1, 1], arr[0, 1], arr[1, 0], arr[0, 0]
        precision, recall, fscore = precision_recall_fscore(tp, fp, fn, beta)
        support = tp + fn
        records.append((class_, precision, recall, fscore, support, tp, fp, fn, tn))

    columns = [
        "class", "precision", "recall", "fscore", "support", "TP", "FP", "FN", "TN"
    ]
    df = pd.DataFrame(
        records,
        columns=columns
    ).sort_values(["support", "TP"], ascending=False)

    support, tp, fp, fn, tn = df[["support", "TP", "FP", "FN", "TN"]].sum(axis=0)
    precision, recall, fscore = precision_recall_fscore(tp, fp, fn, beta)
    return df.append(pd.DataFrame(
        [("avg", precision, recall, fscore, support, tp, fp, fn, tn)],
        columns=columns
    ))
